__all__ = (
    '__title__', '__summary__', '__url__', '__version__',
    '__author__', '__email__', '__license__'
)

__title__ = 'scsm'
__summary__ = 'SteamCMD Server Mangaer'
__url__ = 'https://github.com/bubylou/scsm'

__version__ = '0.1.8'

__author__ = 'Nicholas Malcolm'
__email__ = 'bubylou7@gmail.com'

__license__ = 'GPLv3+'
